@@ -42,11 +42,13 @@ import java.io.IOException;
 import java.net.URI;
 import java.util.HashMap;
 import java.util.Map;
+import java.util.Random;
 import java.util.concurrent.CountDownLatch;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicBoolean;
 import java.util.concurrent.atomic.AtomicReference;
 
+import javax.jms.BytesMessage;
 import javax.jms.Connection;
 import javax.jms.ConnectionFactory;
 import javax.jms.ExceptionListener;
@@ -56,19 +58,22 @@ import javax.jms.Message;
 import javax.jms.MessageConsumer;
 import javax.jms.MessageProducer;
 import javax.jms.Queue;
+import javax.jms.ResourceAllocationException;
 import javax.jms.Session;
 
 import org.apache.qpid.jms.JmsConnection;
 import org.apache.qpid.jms.JmsConnectionFactory;
+import org.apache.qpid.jms.JmsConnectionRemotelyClosedException;
 import org.apache.qpid.jms.JmsDefaultConnectionListener;
-import org.apache.qpid.jms.provider.ProviderRedirectedException;
 import org.apache.qpid.jms.provider.amqp.AmqpSupport;
+import org.apache.qpid.jms.provider.exceptions.ProviderConnectionRedirectedException;
 import org.apache.qpid.jms.test.QpidJmsTestCase;
 import org.apache.qpid.jms.test.Wait;
 import org.apache.qpid.jms.test.testpeer.TestAmqpPeer;
 import org.apache.qpid.jms.test.testpeer.basictypes.AmqpError;
 import org.apache.qpid.jms.test.testpeer.basictypes.ConnectionError;
 import org.apache.qpid.jms.test.testpeer.matchers.CoordinatorMatcher;
+import org.apache.qpid.jms.test.testpeer.matchers.sections.TransferPayloadCompositeMatcher;
 import org.apache.qpid.jms.util.MetaDataSupport;
 import org.apache.qpid.proton.amqp.Binary;
 import org.apache.qpid.proton.amqp.Symbol;
@@ -76,6 +81,7 @@ import org.apache.qpid.proton.amqp.UnsignedInteger;
 import org.apache.qpid.proton.amqp.transaction.TxnCapability;
 import org.apache.qpid.proton.engine.impl.AmqpHeader;
 import org.hamcrest.Matcher;
+import org.junit.Ignore;
 import org.junit.Test;
 
 public class ConnectionIntegrationTest extends QpidJmsTestCase {
@@ -258,6 +264,8 @@ public class ConnectionIntegrationTest extends QpidJmsTestCase {
             txCoordinatorMatcher.withCapabilities(arrayContaining(TxnCapability.LOCAL_TXN));
             testPeer.expectSenderAttach(notNullValue(), txCoordinatorMatcher, true, true, false, 0, null, null);
             testPeer.expectDetach(true, false, false);
+            // Expect the AMQP session to be closed due to the JMS session creation failure.
+            testPeer.expectEnd();
 
             try {
                 connection.createSession(true, Session.SESSION_TRANSACTED);
@@ -374,6 +382,61 @@ public class ConnectionIntegrationTest extends QpidJmsTestCase {
         }
     }
 
+    @Test(timeout = 20000)
+    public void testMaxFrameSizeInfluencesOutgoingFrameSize() throws Exception {
+        doMaxFrameSizeInfluencesOutgoingFrameSizeTestImpl(1000, 10001, 11);
+        doMaxFrameSizeInfluencesOutgoingFrameSizeTestImpl(1500, 6001, 5);
+    }
+
+    private void doMaxFrameSizeInfluencesOutgoingFrameSizeTestImpl(int frameSize, int bytesPayloadSize, int numFrames) throws Exception {
+        try (TestAmqpPeer testPeer = new TestAmqpPeer();) {
+            testPeer.expectSaslLayerDisabledConnect(equalTo(UnsignedInteger.valueOf(frameSize)));
+            // Each connection creates a session for managing temporary destinations etc
+            testPeer.expectBegin();
+
+            String uri = "amqp://localhost:" + testPeer.getServerPort() + "?amqp.saslLayer=false&amqp.maxFrameSize=" + frameSize;
+            ConnectionFactory factory = new JmsConnectionFactory(uri);
+            Connection connection = factory.createConnection();
+
+            testPeer.expectBegin();
+            testPeer.expectSenderAttach();
+
+            Session session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);
+            Queue queue = session.createQueue("myQueue");
+            MessageProducer producer = session.createProducer(queue);
+
+            // Expect n-1 transfers of maxFrameSize
+            for (int i = 1; i < numFrames; i++) {
+                testPeer.expectTransfer(frameSize);
+            }
+            // Plus one more of unknown size (framing overhead).
+            testPeer.expectTransfer(0);
+
+            // Send the message
+            byte[] orig = createBytePyload(bytesPayloadSize);
+            BytesMessage message = session.createBytesMessage();
+            message.writeBytes(orig);
+
+            producer.send(message);
+
+            testPeer.expectClose();
+            connection.close();
+
+            testPeer.waitForAllHandlersToComplete(3000);
+        }
+    }
+
+    private static byte[] createBytePyload(int sizeInBytes) {
+        Random rand = new Random(System.currentTimeMillis());
+
+        byte[] payload = new byte[sizeInBytes];
+        for (int i = 0; i < sizeInBytes; i++) {
+            payload[i] = (byte) (64 + 1 + rand.nextInt(9));
+        }
+
+        return payload;
+    }
+
     @Test(timeout = 20000)
     public void testAmqpHostnameSetByDefault() throws Exception {
         doAmqpHostnameTestImpl("localhost", false, equalTo("localhost"));
@@ -483,9 +546,9 @@ public class ConnectionIntegrationTest extends QpidJmsTestCase {
             assertTrue("Connection should report failure", done.await(5, TimeUnit.SECONDS));
 
             assertTrue(asyncError.get() instanceof JMSException);
-            assertTrue(asyncError.get().getCause() instanceof ProviderRedirectedException);
+            assertTrue(asyncError.get().getCause() instanceof ProviderConnectionRedirectedException);
 
-            ProviderRedirectedException redirect = (ProviderRedirectedException) asyncError.get().getCause();
+            ProviderConnectionRedirectedException redirect = (ProviderConnectionRedirectedException) asyncError.get().getCause();
             URI redirectionURI = redirect.getRedirectionURI();
 
             assertNotNull(redirectionURI);
@@ -520,7 +583,7 @@ public class ConnectionIntegrationTest extends QpidJmsTestCase {
             testPeer.waitForAllHandlersToComplete(1000);
             assertTrue("connection never closed.", Wait.waitFor(new Wait.Condition() {
                 @Override
-                public boolean isSatisified() throws Exception {
+                public boolean isSatisfied() throws Exception {
                     return !((JmsConnection) connection).isConnected();
                 }
             }, 10000, 10));
@@ -586,10 +649,50 @@ public class ConnectionIntegrationTest extends QpidJmsTestCase {
             try {
                 producer.send(message);
                 fail("Expected exception to be thrown");
-            } catch (JMSException jmse) {
+            } catch (ResourceAllocationException jmse) {
                 // Expected
                 assertNotNull("Expected exception to have a message", jmse.getMessage());
                 assertTrue("Expected breadcrumb to be present in message", jmse.getMessage().contains(BREAD_CRUMB));
+            } catch (Throwable t) {
+                fail("Caught unexpected exception: " + t);
+            }
+
+            connection.close();
+
+            testPeer.waitForAllHandlersToComplete(3000);
+        }
+    }
+
+    @Test(timeout = 20000)
+    public void  testRemotelyEndConnectionWithSessionWithProducerWithSendWaitingOnOutcome() throws Exception {
+        final String BREAD_CRUMB = "ErrorMessageBreadCrumb";
+
+        try (TestAmqpPeer testPeer = new TestAmqpPeer();) {
+            Connection connection = testFixture.establishConnecton(testPeer);
+
+            testPeer.expectBegin();
+            Session session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);
+
+            // Expect producer creation, and a message to be sent, but don't return a disposition.
+            // Instead, close the connection.
+            testPeer.expectSenderAttach();
+            testPeer.expectTransferButDoNotRespond(new TransferPayloadCompositeMatcher());
+            testPeer.remotelyCloseConnection(true, ConnectionError.CONNECTION_FORCED, BREAD_CRUMB, 50);
+
+            Queue queue = session.createQueue("myQueue");
+            final MessageProducer producer = session.createProducer(queue);
+
+            Message message = session.createTextMessage("myMessage");
+
+            try {
+                producer.send(message);
+                fail("Expected exception to be thrown");
+            } catch (JmsConnectionRemotelyClosedException jmse) {
+                // Expected
+                assertNotNull("Expected exception to have a message", jmse.getMessage());
+                assertTrue("Expected breadcrumb to be present in message", jmse.getMessage().contains(BREAD_CRUMB));
+            } catch (Throwable t) {
+                fail("Caught unexpected exception: " + t);
             }
 
             connection.close();
@@ -766,4 +869,29 @@ public class ConnectionIntegrationTest extends QpidJmsTestCase {
             testPeer.waitForAllHandlersToComplete(2000);
         }
     }
+
+    @Ignore("Disabled due to requirement of hard coded port")
+    @Test(timeout = 20000)
+    public void testLocalPortOption() throws Exception {
+        try (TestAmqpPeer testPeer = new TestAmqpPeer();) {
+            testPeer.expectSaslAnonymous();
+            testPeer.expectOpen();
+            // Each connection creates a session for managing temporary destinations etc
+            testPeer.expectBegin();
+
+            int localPort = 5671;
+            String uri = "amqp://localhost:" + testPeer.getServerPort() + "?transport.localPort=" + localPort;
+            ConnectionFactory factory = new JmsConnectionFactory(uri);
+            Connection connection = factory.createConnection();
+            connection.start();
+
+            testPeer.waitForAllHandlersToComplete(2000);
+
+            testPeer.expectClose();
+            connection.close();
+
+            int clientPort = testPeer.getClientSocket().getPort();
+            assertEquals(localPort, clientPort);
+        }
+    }
 }
